#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/TADS2/mac/OS0TD_MA.C,v 1.2 1999/05/17 02:52:18 MJRoberts Exp $";
#endif

/* Copyright (c) 1992 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  os0td.c - os mainline for tads2 debugger (Mac version)
Function
  invokes debugger from finder, after allowing user to enter
  parameters through a dialog
Notes
  none
Modified
  08/22/92 MJRoberts     - creation
*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <Types.h>
//#include <OSUtils.h>
#include <ToolUtils.h>
#include <Fonts.h>
#include <Scrap.h>
#include <Dialogs.h>
#include <Files.h>
#include <MacWindows.h>
#include <StandardFile.h>
#include <Lists.h>
#include <Resources.h>
//#include <SegLoad.h>
#include <Menus.h>
#include <Events.h>
#include <TextEdit.h>
#include <Devices.h>
#include <Sound.h>

#include "linf.h"
#include "ossmac.h"
#include "cmd.h"
#include "argize.h"
#include "configs.h"
#include "trd.h"


/* indicates we're on the way out */
int os0td_quitting;

/* covers for malloc and free */
void *osmalloc(size_t siz) { return(malloc(siz)); }
void osfree(void *p) { free(p); }

extern WindowPtr myWindow;

#define INCBUFSIZ 2048

static int DoCommand(long mResult);

/* resource IDs of menus */
#define appleID                 128
#define fileID                  129
#define editID                  130
#define execID                  131
#define dataID                  132
#define breakID                 133
#define sourceID                134
#define watchID                 135

/* Edit menu command indices */
#define undoCommand             1
#define cutCommand              3
#define copyCommand             4
#define pasteCommand    5
#define clearCommand    6

/* Menu indices */
#define appleM                  0
#define fileM                   1
#define editM                   2
#define execM                   3
#define dataM                   4
#define breakM                  5
#define sourceM                 6
#define watchM                  7

#define fmQuit                  1

#define emGo                    1
#define emStep                  2
#define emOver                  3

#define dmEval                  1
#define dmStack                 2
#define dmUp                    3
#define dmDown                  4

#define bmSet                   1
#define bmClear                 2
#define bmDisable               3
#define bmEnable                4

extern MenuHandle myMenus[];

static DialogPtr        globdlg;
static ListHandle       globlst;
static DialogPtr        progdlg;  /* progress dialog */
static DialogPtr        memoryDlg;  /* memory options dialog */

static char                     safetylevel = '0';
static Str255           savefile = "\p";

static struct
{
        unsigned int    setting;
        unsigned int    standard;
        char                    switchar;
//      char                    modified;
} mem_vals[] =
{
        {TDD_POOLSIZ /*6192*/, TDD_POOLSIZ, 'p'},
        {TDD_LCLSIZ  /*2048*/, TDD_LCLSIZ,  'l'}, /* -ml is ignored by tdd.c! */
        {TDD_STKSIZ    /*50*/, TDD_STKSIZ,  's'},
        {TDD_HEAPSIZ /*1024*/, TDD_HEAPSIZ, 'h'}
};
/* XXX: There were some major discrepancies between the default memory values
        in os0td_ma.c and tdd.c.  The dialog will now display the tdd values, which
        are the ones that were used anyway.  The old os0td_ma values are in comments.
*/
static int                      undosize = TDD_UNDOSIZ;


static void get_memory_settings(void)
{
        Str255                  buf;
        char                    bufc[256];
        short                   i;
        unsigned int    newval;
        
        if (!memoryDlg) return;
        
        for (i = 0 ; i < 4 ; ++i)
        {
                GetControlItemTitle(memoryDlg, (i+1)*2, buf);
                StringCopy_cp(bufc, buf, 256);
                newval = (unsigned int) atoi(bufc);
                
                if (newval != mem_vals[i].setting)
                {
                        mem_vals[i].setting = newval;
                }
        }
}

static void set_memory_settings(void)
{
        Str255  buf;
        int             i;
        
        if (!memoryDlg) return;
        
        for (i = 0; i < 4; ++i) {
                psprintf(buf, "%u", mem_vals[i].setting);
                SetControlItemTitle(memoryDlg, (i+1)*2, buf);
        }
}


static void readcommandline(char *buf, ListHandle mylist, DialogPtr mydlg)
{
        #define MAX_ARGS  64
        
        int             argc;
        char   *argv[MAX_ARGS], **argp;
        int             i;
        
        argize(buf, &argc, argv, MAX_ARGS);
        
        /* Set controls to default states. */
        LDelRow(0, (*mylist)->dataBounds.top, mylist);  // Include directory.
        SetControlItemValue(mydlg, 10, 1);                              // Swapping.
        SetControlItemTitle(mydlg, 11, "\p<Default>");
        safetylevel = '0';
        StringCopy_pp(savefile, "\p", sizeof(savefile));
        undosize = TDD_UNDOSIZ;
        
        for (i = 0, argp = argv; i < argc; i++, argp++) {
                char   *arg;
                size_t  arglen;
                
                arg = *argp;
                if (arg[0] == '\0') {
                        /* Someone screwed up and gave us an empty argument -- skip it. */
                }
                else if (arg[0] == '-') {
                        switch(arg[1]) {
                                case 'r': /* Saved game file. */
                                        arg = cmdarg(NULL, &argp, &i, argc, 1, NULL);
                                        StringCopy_pc(savefile, arg, sizeof(savefile));
                                        break;
                                
                                case 'i': /* Include directory. */
                                        arg = cmdarg(NULL, &argp, &i, argc, 1, NULL);
                                        arglen = strlen(arg);
                                        if (arglen) {
                                                Cell    c = {0, 0};
                                                
                                                c.v = LAddRow(1, 32767, mylist);
                                                LSetCell(arg, arglen, c, mylist);
                                        }
                                        break;
                                
                                case 's': /* File safety level. */
                                        arg = cmdarg(NULL, &argp, &i, argc, 1, NULL);
                                        if (*arg) safetylevel = *arg;
                                        break;
                                        
                                case 't': /* Swapping state. */
                                        if (arg[2] == '-') {
                                                SetControlItemValue(mydlg, 10, 0);
                                                SetControlItemTitle(mydlg, 11, "\p");
                                        }
                                        else if (arg[2] == '+') {
                                                SetControlItemValue(mydlg, 10, 1);
                                                SetControlItemTitle(mydlg, 11, "\p<Default>");
                                        }
                                        else if (arg[2] == 'f') {
                                                arg = cmdarg(NULL, &argp, &i, argc, 2, NULL);
                                                if (arg[0]) {
                                                        Str255  swapname;
                                                        
                                                        StringCopy_pc(swapname, arg, sizeof(swapname));
                                                        SetControlItemValue(mydlg, 10, 1);
                                                        SetControlItemTitle(mydlg, 11, swapname);
                                                }
                                        }
                                        break;
                                
                                case 'm': /* Memory settings. */
                                        switch(arg[2]) {
                                                case 'p': mem_vals[0].setting = atoi(cmdarg(NULL, &argp, &i, argc, 2, NULL)); break;
                                                case 'l': mem_vals[1].setting = atoi(cmdarg(NULL, &argp, &i, argc, 2, NULL)); break;
                                                case 's': mem_vals[2].setting = atoi(cmdarg(NULL, &argp, &i, argc, 2, NULL)); break;
                                                case 'h': mem_vals[3].setting = atoi(cmdarg(NULL, &argp, &i, argc, 2, NULL)); break;
                                                default: {
                                                        Str255  cachesize;
                                                        
                                                        arg = cmdarg(NULL, &argp, &i, argc, 1, NULL);
                                                        StringCopy_pc(cachesize, arg, sizeof(cachesize));
                                                        SetControlItemTitle(mydlg, 15, cachesize);
                                                        break;
                                                }
                                        }
                                        break;
                                
                                case 'u':
                                        undosize = atoi(cmdarg(NULL, &argp, &i, argc, 1, NULL));
                                        break;
                        }               
                }
                else {
                        /* If it's not argv[0], and it doesn't begin with '-', it must be the game file. */
                        if (i != 0) {
                                Str255          gamename;
                                
                                StringCopy_pc(gamename, arg, sizeof(gamename));
                                SetControlItemTitle(mydlg, 1, gamename);
                                
                                /* Enable the "Run" button if the name isn't empty. */
                                if (StrLength(gamename) > 0) {
                                        EnableControlItem(mydlg, 8);
                                }
                        }
                }
        }
        
        set_memory_settings();
}
                
static bool writecommandline(DialogPtr mydlg, ListHandle mylist, writeoptions_t &cmdline, bool quote)
{
        #define P(s)  s[0], &s[1]
        
        Str255  itemText;
        Cell    cell;
        int     i;
        
        cmdline.write("tadsdb");
        
        /* Saved game file. */
        if (StrLength(savefile) > 0)
                cmdline.write(quote ? "-r\"%.*s\"" : "-r%.*s", P(savefile));
        
        /* Include directory. */
        cell.v = cell.h = 0;
        while (cell.v < (*mylist)->dataBounds.bottom)
        {
                char   includefile[256];
                short  siz;
                
                siz = 256;
                LGetCell(includefile, &siz, cell, mylist);
                cmdline.write(quote ? "-i\"%.*s\"" : "-i%.*s", siz, includefile);
                cell.v++;
        }
        
        /* File safety level. */
        if (safetylevel != '0')
                cmdline.write("-s%c", safetylevel);
        
        /* Swapping state. */
        if (GetControlItemValue(mydlg, 10)) {
                /* swapping is enabled */
                Str255  swapname;
                
                GetControlItemTitle(mydlg, 11, swapname);
                /* if it's "<Default>", just use -t+ option */
                if (StringEqual_pc(swapname, "<Default>"))
                        // This is the default state -- write nothing.
                        ; //cmdline.write("-t+");
                else
                        cmdline.write(quote ? "-tf\"%.*s\"" : "-tf%.*s", P(swapname));
        }
        else {
                /* swapping is disabled */
                cmdline.write("-t-");
        }
        
        /* Memory settings. */
        get_memory_settings();
        for (i = 0; i < 4; ++i) {
                if (mem_vals[i].setting != mem_vals[i].standard)
                        cmdline.write("-m%c%u", mem_vals[i].switchar, mem_vals[i].setting);
        }
        
        // max cache size.
        GetControlItemTitle(mydlg, 15, itemText);
        if (StrLength(itemText))
                cmdline.write("-m%.*s", P(itemText));
        
        // undo buffer size.
        if (undosize != TDD_UNDOSIZ)
                cmdline.write("-u%d", undosize);
        
        /* Add filename; return whether or not there's a valid game name. */
        GetControlItemTitle(mydlg, 1, itemText);
        if (StrLength(itemText)) {
                cmdline.write(quote ? "\"%.*s\"" : "%.*s", P(itemText));
                return true;
        }
        else
                return false;
        
        #undef P
}


static int getconfig(const char *fname, ListHandle mylist, DialogPtr mydlg)
{
        char            buf[LINE_BUF_SIZE];
        int                     err;
        
        if ((err = readconfig(fname, "tdb", "tadsdb", buf, LINE_BUF_SIZE)) == noErr)
                readcommandline(buf, mylist, mydlg);
        
        return err;
}

static int putconfig(const char *fname, DialogPtr mydlg, ListHandle mylist)
{
        char    buf[LINE_BUF_SIZE];
        
        writecommandline(mydlg, mylist, wrtoptstring_t(buf), true);
        return writeconfig(fname, buf);
}


int topEvent(DialogPtr mydlg, ListHandle mylist)
{
        EventRecord             myEvent;
        WindowPtr               whichWindow;
        int                             gotevent;
        Str255                  gamename;
        Cell                    cell;
        Str255                  itemText;
        int                             oldval;
        static int              first_time = 1;
        
        if (!mydlg) mydlg = globdlg;
        if (!mylist) mylist = globlst;
        
        /*
           maintain buttons
        */
        
        /* turn on "Remove" button only if something in list is hilited */
        cell.v = cell.h = 0;
        HiliteControlItem(mydlg, 7, LGetSelect(TRUE, &cell, mylist) ? 0 : 255);
        
        /* likewise "Select (swap file)" */
        oldval = GetControlItemValue(mydlg, 10);
        HiliteControlItem(mydlg, 13, oldval ? 0 : 255);
        
        /* and "Default" (swap file) button */
        HiliteControlItem(mydlg, 18, oldval ? 0 : 255);
        
        SystemTask();
        if (first_time)
        {
                /* select game file on entry if no game file is selected */
                first_time = 0;
                GetControlItemTitle(mydlg, 1, itemText);
                if (itemText[0] == 0) goto select_game;
        }
        
        gotevent = WaitNextEvent(everyEvent, &myEvent, 0L, 0L);
        if (IsDialogEvent(&myEvent))
        {
                DialogPtr       dlg;
                short           itemHit;
                char            incbuf[INCBUFSIZ];
                Point           oldwhere;
                
                oldwhere = myEvent.where;
                
                if (myEvent.what == updateEvt && (WindowPtr)myEvent.message == (WindowPtr)mydlg)
                {
                        LUpdate(((GrafPtr)mydlg)->visRgn, mylist);
                        SetPort((GrafPtr)mydlg);

                        drawbox(mydlg, 1);      /* draw box around source filename */
                        drawbox(mydlg, 11);     /* draw box around swap filename */
                        drawbox(mydlg, 5);      /* draw box around include list */
                }

                if (DialogSelect(&myEvent, &dlg, &itemHit))
                {
                        if (dlg == memoryDlg)
                        {
                                return(1);
                        }
                        
                        if (dlg == mydlg)
                        {
                                switch(itemHit)
                                {
                                        case 17:   /* "Memory..." button */
                                                if (memoryDlg)
                                                        SelectWindow((WindowPtr)memoryDlg);
                                                else
                                                {
                                                        memoryDlg = GetNewDialog(13919, NULL, (WindowPtr) -1L);
                                                        set_memory_settings();
                                                }
                                                break;

                                        case 3:   /* "Select" button - select game name */
                                        select_game:
                                                if (os0askfile("Game Source File", gamename, 'TADG', 0))
                                                        break;
                                                SetControlItemTitle(mydlg, 1, gamename);
                                                
                                                /* we can turn on the "OK" button now */ 
                                                EnableControlItem(mydlg, 8);
                                                break;
                                        
                                        case 6:   /* "Add" button - add include folder */
                                                if (osaskdir(incbuf, INCBUFSIZ))
                                                        break;
                                                cell.v = LAddRow(1, 32767, mylist);
                                                cell.h = 0;
                                                LSetCell(incbuf, (int)strlen(incbuf), cell, mylist);
                                                LDraw(cell, mylist);
                                                break;
                                        
                                        case 7:   /* "Remove" button - remove an include folder */
                                                cell.v = cell.h = 0;
                                                while (LGetSelect(TRUE, &cell, mylist))
                                                        LDelRow(1, cell.v, mylist);
                                                break;
                                        
                                        case 8:   /* "OK" button */ {
                                                char               *argv[50];
                                                int                             argc = 0;
                                                
                                                if (!writecommandline(mydlg, mylist, wrtoptargv_t(incbuf, argc, argv), false))
                                                        break;
                                                
                                                /* save the dialog state */
                                                putconfig(NULL, mydlg, mylist);
                                                
                                                /* get rid of the dialog - it's no longer needed */
                                                DisposeDialog(mydlg);
                                                if (memoryDlg) DisposeDialog(memoryDlg);
                                                
                                                /* enter the debugger */
                                                SetUpWindows();
                                                os0main2(argc, argv, tddmain, "", ""/*"config.tdb"*/, NULL);
                                                
                                                /* that's it - terminate the program */
                                                {
                                                        extern WindowPtr watchWindow;
                                                        if (watchWindow)
                                                                DisposeWindow(watchWindow);
                                                        watchWindow = NULL;     /* make sure we don't update watches now */
                                                }
                                                os0td_quitting = 1;
                                                os_expause();
                                                return 0;
                                        }
                                        case 9:  /* "Quit" button */
                                                putconfig(NULL, mydlg, mylist);
                                                return 0;
                                                
                                        case 5:  /* include file list box */
                                                SetPort((WindowPtr)mydlg);
                                                GlobalToLocal(&oldwhere);
                                                LClick(oldwhere, myEvent.modifiers, mylist);
                                                break;
                                                
                                        case 10:  /* "Allow Swapping" check box */
                                                ToggleControlItemValue(mydlg, itemHit);
                                                break;
                                        
                                        case 13:  /* "Select" (swap file) box */
                                                if (os0askfile("Swap File", gamename, 'TEXT', 1))
                                                        break;
                                                SetControlItemTitle(mydlg, 11, gamename);
                                                break;
                                
                                        case 18:  /* Default swap file button */
                                                SetControlItemTitle(mydlg, 11, "\p<Default>");
                                                break;
                                }
                        }
                }
                
                return(1);
        }
        
        if (gotevent)
        {
                switch (myEvent.what)
                {
                        case mouseDown:
                                switch (FindWindow( myEvent.where, &whichWindow ))
                                {
                                        case inDesk: 
                                                SysBeep(10);
                                                break;
                                                
                                        case inGoAway:
                                                if (!TrackGoAway(whichWindow, myEvent.where)) break;
                                                if (memoryDlg && whichWindow == (WindowPtr)memoryDlg)
                                                {
                                                                get_memory_settings();
                                                                DisposeDialog(memoryDlg);
                                                                memoryDlg = 0;
                                                }
                                                break;

                                        case inMenuBar:
                                                return(DoCommand(MenuSelect(myEvent.where)));

                                        case inSysWindow:
                                                SystemClick( &myEvent, whichWindow );
                                                break;
                                                
                                        case inDrag:
                                                if (ours(whichWindow) || whichWindow == (WindowPtr)mydlg
                                                        || (progdlg && whichWindow == (WindowPtr)progdlg)
                                                        || (memoryDlg && whichWindow == (WindowPtr)memoryDlg))
                                                {
                                                        static Rect dragRect = { 0, 0, 1024, 1024 };
                                                        
                                                        /* oldfront = whichWindow; */
                                                        DragWindow(whichWindow, myEvent.where, &dragRect);
                                                }
                                                break;
                                                
                                        case inContent:
                                                if (whichWindow != FrontWindow())
                                                {
                                                        SelectWindow(whichWindow);
                                                        /* oldfront = whichWindow; */
                                                }
                                                else if (whichWindow == (WindowPtr)mydlg)
                                                {
                                                        LClick(myEvent.where, myEvent.modifiers, mylist);
                                                }
                                                else if (ours(whichWindow))
                                                        return(DoContent(whichWindow, &myEvent));
                                                break;
                                                
                                        case inGrow:
                                                if (ours(whichWindow)) MyGrowWindow(whichWindow, myEvent.where);
                                                break;

                                        default:
                                                break;
                                }
                                break;
                                        
                        case activateEvt:
                        {
                                Rect r;
                                extern TEHandle TEH;
                                extern ControlHandle vScroll;

                                if (ours(whichWindow = (WindowPtr)myEvent.message)) {
                                        r = whichWindow->portRect;
                                        r.top = r.bottom - (kScrollBarWidth+1);
                                        r.left = r.left - (kScrollBarWidth+1);
                                        InvalRect(&r);
                                        if ( myEvent.modifiers & activeFlag ) {
                                                TEActivate( TEH );
                                                ShowControl( vScroll );
                                                TEFromScrap();
                                        }
                                        else {
                                                TEDeactivate(TEH);
                                                HideControl( vScroll );
                                                ZeroScrap();
                                                TEToScrap();
                                        }
                                }
                                break;
                        }

                        case updateEvt: 
                                if (ours(whichWindow = (WindowPtr)myEvent.message))
                                        UpdateWindow(whichWindow);
                                break;
                                
                        case keyDown:
                        case autoKey:
                        {
                                char theChar = (myEvent.message & charCodeMask);
                                
                                if ((myEvent.modifiers & cmdKey) && theChar == '.') return(2);
                                if (myEvent.modifiers & cmdKey)
                                        return(DoCommand(MenuKey(theChar)));
                                break;
                        }

                        default:
                                break;
                }
        }
        
        return(1);
}

static void SetUpMenus(void)
{
        int i;
        
        myMenus[appleM] = GetMenu(appleID);
        AppendResMenu(myMenus[appleM], 'DRVR');
        
        myMenus[fileM] = GetMenu(fileID);
        myMenus[editM] = GetMenu(editID);
        myMenus[execM] = GetMenu(execID);
        myMenus[dataM] = GetMenu(dataID);
        myMenus[breakM] = GetMenu(breakID);
        myMenus[sourceM] = GetMenu(sourceID);
        myMenus[watchM] = GetMenu(watchID);

        for (i = appleM ; i <= watchM ; i++)
                InsertMenu(myMenus[i], 0) ;
        DrawMenuBar();

        DisableItem(myMenus[editM], undoCommand);
        DisableItem(myMenus[editM], cutCommand);
        DisableItem(myMenus[editM], copyCommand);
        DisableItem(myMenus[editM], clearCommand);
        DisableItem(myMenus[editM], pasteCommand);
        DisableItem(myMenus[editM], 8);
        DisableItem(myMenus[editM], 9);
        
        /* disable all debugging menus for now */
        DisableItem(myMenus[execM], 1);
        DisableItem(myMenus[execM], 2);
        DisableItem(myMenus[execM], 3);
        DisableItem(myMenus[execM], 4);
        DisableItem(myMenus[execM], 5);
        DisableItem(myMenus[execM], 6);
        DisableItem(myMenus[execM], 7);
        DisableItem(myMenus[dataM], 1);
        DisableItem(myMenus[dataM], 2);
        DisableItem(myMenus[dataM], 3);
        DisableItem(myMenus[dataM], 4);
        DisableItem(myMenus[breakM], 1);
        DisableItem(myMenus[breakM], 2);
        DisableItem(myMenus[breakM], 3);
        DisableItem(myMenus[breakM], 4);
        DisableItem(myMenus[breakM], 5);
        DisableItem(myMenus[sourceM], 1);
        DisableItem(myMenus[sourceM], 2);
        DisableItem(myMenus[sourceM], 3);
        DisableItem(myMenus[watchM], 1);
        DisableItem(myMenus[watchM], 2);
}

static int DoCommand(long mResult)
{
        short   theItem;
        Str255  name;
        char   *newcmd = NULL;
        long    cmdlen = 0;
        
        theItem = LoWord(mResult);
        switch (HiWord(mResult))
        {
                case appleID:
                        if (theItem == 1) /* about */
                        {
                                DialogPtr       mydlg;
                                short           itemHit;
                                
                                mydlg = GetNewDialog(7768, NULL, (WindowPtr) -1L);
                                do
                                {
                                        ModalDialog(NULL, &itemHit);
                                } while (itemHit != 1);
                                DisposeDialog(mydlg);
                        }
                        else
                        {
                                GetMenuItemText(myMenus[appleM], theItem, name);
                                OpenDeskAcc(name);
                        }
                        break;
                
                case fileID:
                        switch( theItem )
                        {
                                case fmQuit:
                                        return 0;
                        }
                        break;
        }

        HiliteMenu(0);
        return 1;
}


static OSErr InitToolbox__noAE(void)
{
        extern short    gAppResFile;
        
        OSErr   err = noErr;
        
        #if DEBUG
        static bool     done_init = false;
        
        assert(!done_init);
        done_init = true;
        #endif
        
        SetApplLimit(GetApplLimit() - (48L * 1024L));
        MaxApplZone();
        
        InitGraf(&qd.thePort);
        InitFonts();
        InitWindows();
        InitMenus();
        TEInit();
        InitDialogs(NULL);
        InitCursor();
        FlushEvents(everyEvent, 0);

        gAppResFile = CurResFile();
        
        return err;
}


void main()
{
        DialogPtr       mydlg;
        Str255          itemText;
        Rect            itemRect;
        ListHandle      mylist;
        Rect            dataBounds;
        Point           pt;
        Cell            cell;
        FSSpec          appFSSpec;
        char            apppath[256];

        if (InitToolbox__noAE() == noErr) {
                /* do os initialization */
                os_init(0, (char **)0, (char *)0, (char *)0, 0);
                SetUpMenus();
                
                /* get current directory (in string format) */
                (void) GetApplicationInfo(NULL, &appFSSpec);
                dirtopath(appFSSpec.vRefNum, appFSSpec.parID, apppath);
                
                /* set up dialog and include file list box */
                mydlg = GetNewDialog(19587, NULL, (WindowPtr) -1L);
                GetControlItemRect(mydlg, 5, &itemRect);
                itemRect.right -= 15;  /* for scroll bar */
                dataBounds.top = 0; dataBounds.left = 0; dataBounds.bottom = 0;
                dataBounds.right = 0;
                pt.h = pt.v = 0;
                mylist = LNew(&itemRect, &dataBounds, pt, 0, (WindowPtr)mydlg, TRUE, FALSE, FALSE, TRUE);
                LAddColumn(1, 0, mylist);
                
                if (getconfig(NULL, mylist, mydlg) != noErr)
                {
                        /* add a row with the current directory to the include path */
                        cell.v = LAddRow(1, 32767, mylist);
                        cell.h = 0;
                        LSetCell(apppath, strlen(apppath), cell, mylist);
                
                        /* turn on swapping by default */
                        SetControlItemValue(mydlg, 10, 1);
                        SetControlItemTitle(mydlg, 11, "\p<Default>");
                }
                
                /* make "OK" button inactive until we get a filename */
                GetControlItemTitle(mydlg, 1, itemText);
                HiliteControlItem(mydlg, 8, itemText[0] ? 0 : 255);
        
                while (topEvent(mydlg, mylist))
                        ;
                
                /* os_expause(); */
        }
        XXX_FixStandAlone();
}

