#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/TADS2/mac/OSSMAC.C,v 1.2 1999/05/17 02:52:18 MJRoberts Exp $";
#endif

/* Copyright (c) 1992 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  ossmac.c  - macintosh os support routines
Function
  Routines to implement text windows for Mac version of TADS
Notes
  These routines have been separated out from osmac.c mostly
  because combining them would result in a really huge file.
  The routines in this file are functionally related:  they
  handle the event loop and window updating.
  
  Note that some externals are in TDB 2.0; when building the
  compiler or run-time, these externals should simply be
  stubbed out, since they won't be called but are needed for
  linking.
Modified
  01/20/91 MJRoberts   - created
*/

#include "ossmac.h"

#include <stdlib.h>
#include <stddef.h>
#include <string.h>

#include <QuickDraw.h>
//#include <OSUtils.h>
#include <ToolUtils.h>
#include <Fonts.h>
#include <Scrap.h>
#include <Types.h>
#include <Files.h>
#include <Windows.h>
#include <Fonts.h>
#include <TextEdit.h>
#include <Dialogs.h>
#include <Events.h>
#include <Controls.h>
#include <ToolUtils.h>
#include <Menus.h>
#include <Devices.h>
#include <AppleEvents.h>
#include <Sound.h>
#include <NumberFormatting.h>
#include <LowMem.h>

#include "run.h"
#include "dbgumac.h"
#include "rtprefs.h"

#define qdPort (qd.thePort)
#define qdArrow (qd.arrow)


#define STATUS_OFFSET 18

/* resource IDs of menus */
#define appleID                 128
#define fileID                  129
#define editID                  130
#define cmdID                   131
#define travelID                132
#define fontID                  133
#define fontsizeID              134

#define windowID                128
#define dbgWinID                129

/* Edit menu command indices */
#define undoCommand             1
#define cutCommand              3
#define copyCommand             4
#define pasteCommand    5
#define clearCommand    6

/* Menu indices */
#define appleM                  0
#define fileM                   1
#define editM                   2
#define cmdM                    3
#define travelM                 4
#define fontM                   5
#define fontsizeM               6

MenuHandle                              myMenus[10];

#define fmSave                  1
#define fmRestore               2
#define fmRestart               3
#define fmQuit                  5

#define cmInventory             1
#define cmLook                  2
#define cmScore                 3
#define cmWait                  4
#define cmTake                  6
#define cmDrop                  7

#define cmNorth                 1
#define cmSouth                 2
#define cmEast                  3
#define cmWest                  4
#define cmNE                    5
#define cmNW                    6
#define cmSE                    7
#define cmSW                    8
#define cmUp                    9
#define cmDown                  10
#define cmIn                    11
#define cmOut                   12


#define aaSave                  1
#define aaDiscard               2
#define aaCancel                3

WindowRecord    wRecord;
WindowPtr               myWindow;
WindowPtr               dbgWindow, gameWindow, srcWindow, watchWindow;
TEHandle                TEH, dTEH;
int                             linesInFolder;
Rect                    dragRect = { 0, 0, 1024, 1024 };
ControlHandle   vScroll;
ControlHandle   dvScroll, svScroll, shScroll;
ControlHandle   wvScroll, whScroll;
WindowRecord    dbgWRecord;
WindowRecord    srcWRecord;
Cursor                  editCursor;
Cursor                  waitCursor;
extern int              tadsdebug, tdb2;

static bool             gsQuitting = false;


struct btndef
{
        unsigned char  *btnname;
        char                    btnrefcon;
        ControlHandle   btnctl;
};
typedef struct btndef btndef;

btndef btnlist[] =
{
        { "\pStep",    ' ' },
        { "\pGo",      'g' },
        { "\pReturn",  'r' },
/*      { "\pAbort",   'a' }, */  /* not enough space - remove button */
        { "\pTrace",   't' },
/*      { "\pWhere",   'w' }, */  /* not enough space - remove button */
        { "\pEval",    'e' },
        { "\pSet Brk", 'b' },
        { "\pDel Brk", 'd' },
        { "\pList Brk",'l' },
        { 0,            0  }
};


/* forward declarations */
void MaintainCursor(void);
void MaintainMenus(void);
void SetUpCursors(void);
void AdjustText(WindowPtr w);
void SetControlEnd(void);
void ShowEnd(void);
void SetVScroll(void);
void ShowSelect(void);
void SetView(WindowPtr w, bool recalcAll = false);
int DoCommand(long mResult);
int MainEvent(void);


static void TESetEmptySelection(TEHandle te)
{
        TESetSelect(32767L, 32767L, te);
}

static void findinfo(WindowPtr w, TEHandle *te, ControlHandle *vs)
{
        if (w == gameWindow)
        {
                *te = TEH;
                *vs = vScroll;
        }
        else if (w == dbgWindow)
        {
                *te = dTEH;
                *vs = dvScroll;
        }
        else if (w == srcWindow)
        {
                *te = 0;
                *vs = svScroll;
        }
        else if (w == watchWindow)
        {
                *te = 0;
                *vs = wvScroll;
        }
        else
        {
                *te = NULL;
                *vs = NULL;
        }
}

void editinit(void) 
{
        #if 0
        InitGraf(&qdPort);
        InitFonts();
        FlushEvents(everyEvent, 0);
        InitWindows();
        InitMenus();
        TEInit();
        InitDialogs(0L);
        InitCursor();
        MaxApplZone();
        #endif

        SetFractEnable(true);
        SetUpCursors();
        if (!tdb2) SetUpWindows();
}

void os_update(void)
{
        UpdateWindow(myWindow);
        if (dbgWindow) UpdateWindow(dbgWindow);
}

void os_flush(void)
{
        /* ShowSelect(); */
}

static int firstline = 0; /* number of lines in TEH at last input */

void oscls(void)
{
        TESetSelect(0L, 32767L, TEH);
        TEDelete(TEH);
        SetControlEnd();
        ShowSelect();
        AdjustText(myWindow);
}

void os_more_prompt()
{
        long            end;
        TEHandle        te = (myWindow == dbgWindow ? dTEH : TEH);
        
        TEInsert("\r* * * More * * *", 17L, te);
        os_waitc();
        end = (*te)->teLength;
        TESetSelect(end - 17, end, te);
        TEDelete(te);
        TESetEmptySelection(te);
        firstline = (*te)->nLines;
}

void os_putln(char *s)
{
        #define MAXBUFSIZE (24*1024)
                // Was 30.
        #define REDUCESIZE (4*1024)
                // Was 5.
        
        #if DEBUG
        static char        *gsp, *gss;
        static long             gsLen;
        static short    gsTELength;
        #endif
        
        char       *p;
        long            len;
        char            c;
        TEHandle        te = (myWindow == dbgWindow ? dTEH : TEH);
        
        if (myWindow != FrontWindow())
                SelectWindow(myWindow);

        TESetEmptySelection(te);
        for (p = s; *s; )
        {
                len = 0;
                while (*s && *s != '\n' && *s != '\r') ++s, ++len;
                c = *s;
                *s = '\0';
                #if DEBUG
                // If TEInsert dies, the debugger tends to lose track of local
                // variables, so save them now.
                gsp = p;
                gss = s;
                gsLen = len;
                gsTELength = (*te)->teLength;
                #endif
                TEInsert(p, len, te);
                if (te == TEH)
                {
                        extern int moremode;
                        int displaylines;
                        
                        displaylines = Dy((**te).viewRect) / (**te).lineHeight;
                        
                        if (moremode && (*te)->nLines - firstline + 3 > displaylines) {
                                #if DEBUG
                                int lines_over = (*te)->nLines - firstline + 3 - displaylines;
                                #endif
                                os_more_prompt();
                        }
                }
                TESetEmptySelection(te);
                if (c == '\n')
                        TEKey(kCR, te);
                else if (c == '\r')
                {
                        /*
                         *   Deal with carriage returns:  these are supposed to
                         *   overwrite everything to the beginning of the line.
                         */
                        char *q, *r;

                        DPRINTF(("ossmac::os_putln - got a linefeed!"));
                        q = *TEGetText(te);
                        for (r = q + (*te)->teLength ; r>q && *r != kCR; --r) ;
                        if (*r == kCR) ++r;             /* don't delete the CR if we actually got one */
                        
                        TESetSelect((long)(r-q), (long)((*te)->teLength), te);
                        TEDelete(te);
                        TESetEmptySelection(te);
                }
                *s = c;
                if (*s) p = ++s;
                
                /*
                 *   Make sure the buffer isn't too long.  If it is, cut it down to the limit
                 *   size, and readjust everything to make up for it.
                 */
                if ((*te)->teLength > MAXBUFSIZE)
                {
                        int             i;
                        char   *p;
                        short   oldnlines, lineslost;
                        
                        TEAutoView(false, te);                                  /* turn off automatic selection-showing */
                        oldnlines = (*te)->nLines;
                        p = *((*te)->hText) + REDUCESIZE;
                        i = REDUCESIZE;
                        while (*p && *p != kCR) ++p, ++i;               /* find the end of a line */
                        TESetSelect(0L, (long)i+1, te);                 /* select start of buffer to the found line */
                        TEDelete(te);                                                   /* delete that portion */
                        lineslost = oldnlines - (*te)->nLines;  /* adjust firstline */
                        firstline -= lineslost;
                        TEAutoView(true, te);
                }
        }
        
        ShowEnd();
}

int os_hilite(int flag /* 2 = on, 1 = off */)
{
        #ifdef NEVER
        TextStyle t;
        
        t.tsFace = (flag == 2 ? bold : 0);
        TESetStyle(doFace, &t, 0, TEH);
        return(0);
        #endif /* NEVER */
        return 0;
}

static int dbgyloc;
int dbgydelta;
int dbgpagesize;
#define DBGLINE1 35
#define DBGPTR   0xA5 /* pointer highlighting a source line */

static char             dbgsrcbuf[1024];
static char        *dbgsrcptr = dbgsrcbuf;
static int              dbglin;
static int              dbghilite = -1;
static int              dbghiflag;

static void macedDbUpd(void)
{
}

void macedDbPut(char *s)
{
}

#if 0
/* None of these functions are used anywhere. */
void os_dbhi(int lin, int flag)
{
}

void os_dbgini(void)
{
}

int os_dbpage(void)
{
        return(dbgpagesize);
}

void os_dbcol(int flag)
{
}
#endif


int                             cmdpoint;                                       // Not static -- linked to in dbgumac.c!
static char             onechar, onechar_keycode;

static int              pending_key = 0;

static int interpret_key(char c, char keycode)
{
        switch (c) {
                case kHomeCharCode:                     pending_key = CMD_HOME; break;
                case kEnterCharCode:            return '\n';
                case kEndCharCode:                      pending_key = CMD_END; break;
        //      case kTabCharCode:                      pending_key = CMD_TAB; break;
                case kPageUpCharCode:           pending_key = CMD_PGUP; break;
                case kPageDownCharCode:         pending_key = CMD_PGDN; break;
                case kLeftArrowCharCode:        pending_key = CMD_LEFT; break;
                case kRightArrowCharCode:       pending_key = CMD_RIGHT; break;
                case kUpArrowCharCode:          pending_key = CMD_UP; break;
                case kDownArrowCharCode:        pending_key = CMD_DOWN; break;
                case kDeleteCharCode:           pending_key = CMD_DEL; break;
                case kFunctionKeyCharCode: {
                        switch (keycode) {
                                case 0x7A: pending_key = CMD_F1; break;
                                case 0x78: pending_key = CMD_F2; break;
                                case 0x63: pending_key = CMD_F3; break;
                                case 0x76: pending_key = CMD_F4; break;
                                case 0x60: pending_key = CMD_F5; break;
                                case 0x61: pending_key = CMD_F6; break;
                                case 0x62: pending_key = CMD_F7; break;
                                case 0x64: pending_key = CMD_F8; break;
                                case 0x65: pending_key = CMD_F9; break;
                                case 0x6D: pending_key = CMD_F10; break;
                                
                                default: return c;
                        }
                        break;
                }
                
                default: return c;
        }
        return 0;
}

int os_getc(void)
{
        char            c;
        
        if (pending_key) {
                c = pending_key;
                pending_key = 0;
                return c;
        }
        else {
                TEHandle        te = (myWindow == dbgWindow ? dTEH : TEH);

                SetControlEnd();
                TESetEmptySelection(te);
                ShowSelect();
                cmdpoint = (*te)->selStart;
                onechar = 1;
                if (te == TEH) firstline = (*TEH)->nLines;
                while (MainEvent())
                        ;
                
                return interpret_key(onechar, onechar_keycode);
        }
}

int os_getc_raw(void)
{
        /* 
         *   our key codes are already unambiguous - we never map anything
         *   to a translated CMD_xxx code when it could also be
         *   represented as a raw code - so we can just return the same
         *   thing as os_getc 
         */
        return os_getc();
}

extern int use_statline, run_statline, dbgactive;

unsigned char *os_gets(unsigned char *buf, size_t bufl)
{
        char       *p;
        int                     i;
        char       *dst;
        int                     len;
        TEHandle        te = (myWindow == dbgWindow ? dTEH : TEH);
        extern char     cmdbuf[];
        
        SetControlEnd();
        if (use_statline && run_statline && !dbgactive
                && !(tdb2 && te == dTEH))
                runstat(0);
                
        /*
          if running tdb2, and we have a watch window, and we're playing the game,
          update the watch window to keep it current before each move
        */
        if (watchWindow && myWindow == gameWindow) dbgwxinval();

        /*
         *   Select the end of the buffer, show the selection, save the current location
         *   for later comparison of where the command line begins, and allow normal text
         *   editing.  Keep editing text until the RETURN or ENTER key is hit.
         */
        TESetEmptySelection(te);
/*      ShowSelect(); */
        cmdpoint = (*te)->selStart;

        if (FrontWindow() != myWindow)
                        SelectWindow(myWindow);

        onechar = 0;
        if (te == TEH) firstline = (*te)->nLines;
        SetVScroll();
        Draw1Control(vScroll);
        while (MainEvent())
                ;

        /*
         *   Now copy the text from the start of the command line to the end of the
         *   edit buffer.  Limit the text to bufl-1 characters, so that we have room for
         *   a null byte at the end.
         */
        p = (*((*te)->hText)) + cmdpoint;
        for (len=bufl-1, dst=(char *)buf, i=(*te)->teLength; i>cmdpoint && len; --i, --len)
                *dst++ = *p++;

        *dst++ = '\0';
        TESetEmptySelection(te);
        TEKey(kCR, te);
        return(buf);
}

static void TypeCommand(const char *cmd, TEHandle te)
{
        if (!te) {
                ControlHandle   vs;
                
                findinfo(myWindow, &te, &vs);
                assert(te);
        }
        if (VERIFY(cmd)) {
                size_t  cmdlen = strlen(cmd);
                
                assert(cmdlen > 0);
                TESetSelect((long) cmdpoint, 32767L, te);
                TEDelete(te);
                TEInsert(cmd, (cmdlen > 0 ? cmdlen : -cmdlen), te);
        }
}

void HandleQuitApp()
{
        assert(!gsQuitting);
        gsQuitting = true;
}

int MainEvent(void) 
{
        EventRecord                     myEvent;
        WindowPtr                       whichWindow;
        Rect                            r;
        static WindowPtr        oldfront;
        TEHandle                        te = (myWindow == dbgWindow ? dTEH : TEH);
        ControlHandle           vs = (myWindow == dbgWindow ? dvScroll : vScroll);
        
        assert(!gsQuitting);
        if (!oldfront) oldfront = myWindow;
        if (tadsdebug && oldfront && (oldfront != FrontWindow()))
                SelectWindow(oldfront);
        
        MaintainCursor();
        MaintainMenus();
        SystemTask();
        TEIdle(te);
        if (dTEH) TEIdle(dTEH);
        if (GetNextEvent(everyEvent, &myEvent))
        {
                switch (myEvent.what)
                {
                        case mouseDown:
                                switch (FindWindow(myEvent.where, &whichWindow))
                                {
                                        case inDesk: 
                                                SysBeep(10);
                                                break;
                                                
                                        case inMenuBar:
                                                return(DoCommand(MenuSelect(myEvent.where)));
                                                
                                        case inSysWindow:
                                                SystemClick(&myEvent, whichWindow);
                                                break;
                                                
                                        case inDrag:
                                                if (whichWindow != FrontWindow())
                                                {
                                                        SelectWindow(whichWindow);
                                                        oldfront = whichWindow;
                                                }
                                                else if (ours(whichWindow))
                                                {
                                                        oldfront = whichWindow;
                                                        DragWindow(whichWindow, myEvent.where, &dragRect);
                                                        if (whichWindow == gameWindow) {
                                                                TPortState      portState(whichWindow);
                                                                Rect            bounds;
                                                                
                                                                bounds = whichWindow->portRect;
                                                                LocalToGlobal((Point *) &bounds.top);
                                                                LocalToGlobal((Point *) &bounds.bottom);
                                                                RTPSetWindowUserSize(bounds);
                                                        }
                                                }
                                                break;
                                                
                                        case inGrow:
                                                if (ours(whichWindow))
                                                {
                                                        MyGrowWindow(whichWindow, myEvent.where);
                                                        if (whichWindow == gameWindow)
                                                                firstline = (*TEH)->nLines;
                                                }
                                                break;
                                                
                                        case inContent:
                                                if (whichWindow != FrontWindow())
                                                {
                                                        SelectWindow(whichWindow);
                                                        oldfront = whichWindow;
                                                }
                                                else if (ours(whichWindow))
                                                        return(DoContent(whichWindow, &myEvent));
                                                break;
                                }
                                break;
                                
                        case keyDown:
                        case autoKey:
                        {
                                char    theChar;
                                
                                theChar = myEvent.message & charCodeMask;
                                if ((myEvent.modifiers & cmdKey) != 0) 
                                        return(DoCommand(MenuKey(theChar)));
                                        
                                if (theChar == kBackspaceCharCode && watchWindow && FrontWindow() == watchWindow) {
                                        dbgwxclr();
                                        return(0);  // XXX: A guess... [cdn 10/25/97]
                                }

                                if (myWindow != FrontWindow()) {
                                        SelectWindow(myWindow);
                                        oldfront = myWindow;
                                }
                                
                                if (!onechar &&
                                        theChar != kPageUpCharCode && theChar != kPageDownCharCode &&
                                        theChar != kHomeCharCode && theChar != kEndCharCode)
                                {
                                        /* if it's not a view-move key, force the command line into view. */
                                        if (GetControlValue(vs) != GetControlMaximum(vs)) ShowSelect();
                                        
                                        if ((*te)->selEnd < cmdpoint)   /* whole selection out of command line */
                                                TESetEmptySelection(te);                /* so select end of cmdline */
                                        else if ((*te)->selStart < cmdpoint)    /* starts before command line */
                                                TESetSelect((long)cmdpoint, (long)(*te)->selEnd, te);
                                }
                                
                                if (onechar) {
                                        onechar = theChar;
                                        onechar_keycode = (myEvent.message & keyCodeMask) >> 8;
                                        return(0);
                                }
                                                
                                if (theChar == kBackspaceCharCode && (*te)->selStart == (*te)->selEnd && (*te)->selStart == cmdpoint)
                                        break;          /* ignore backspace at start of command line */
                                else if (theChar == kLeftArrowCharCode && (*te)->selStart == cmdpoint)
                                        break;          /* ignore left arrow at start of command line */
                                else if (theChar == kUpArrowCharCode || theChar == kDownArrowCharCode)
                                        break;          /* ignore up/down arrows at all times */
                                else if (theChar == kReturnCharCode || theChar == kEnterCharCode)
                                        return 0;       /* return key - we're done with this line */
                                else if (theChar == kPageUpCharCode || theChar == kPageDownCharCode) {
                                        int                     pageSize;
                                        int                     scrollAmt;
                                                                                
                                        pageSize = Dy((**te).viewRect) / (**te).lineHeight - 1;
                                        scrollAmt = (theChar == kPageDownCharCode) ? pageSize : -pageSize;
                                        SetControlValue(vs, GetControlValue(vs) + scrollAmt);
                                        AdjustText(myWindow);
                                        break;
                                }
                                else if (theChar == kHomeCharCode) {
                                        SetControlValue(vs, GetControlMinimum(vs));
                                        AdjustText(myWindow);
                                        break;
                                }
                                else if (theChar == kEndCharCode) {
                                        SetControlValue(vs, GetControlMaximum(vs));
                                        AdjustText(myWindow);
                                        break;
                                }
                                
                                if (theChar != kBackspaceCharCode && theChar != kRightArrowCharCode && theChar != kLeftArrowCharCode) {
                                        if (theChar < 0x20)
                                                break;  /* if it's a control character or forward-delete, ignore it. */
                                        
                                        if ((*te)->selStart == (*te)->selEnd && (*te)->teLength - cmdpoint > 126)
                                                break;  /* can't insert any more; the line is too long */
                                }

                                {
                                        int oldnlines = (*te)->nLines;
                                        
                                        if (theChar == kDeleteCharCode) {
                                                /* TextEdit doesn't support forward delete, so we have to do it ourselves. */
                                                if ((**te).selStart == (**te).selEnd)
                                                        TESetSelect((**te).selStart, (**te).selStart+1, te);
                                                TEDelete(te);
                                        }
                                        else
                                                TEKey(theChar, te);
                                        
                                        if ((*te)->nLines != oldnlines) ShowSelect();
                                }
                                break;
                        }
                        
                        case activateEvt:
                                if (ours(whichWindow = (WindowPtr)myEvent.message))
                                {
                                        r = whichWindow->portRect;
                                        r.bottom -= kScrollBarWidth + 1;
                                        r.right -= kScrollBarWidth + 1;
                                        SetPort(whichWindow);
                                        InvalRect(&r);
                                        findinfo(whichWindow, &te, &vs);
                                        if (myEvent.modifiers & activeFlag)
                                        {
                                                if (te) TEActivate(te);
                                                if (vs) ShowControl(vs);
                                                if (whichWindow == srcWindow) ShowControl(shScroll);
                                                if (whichWindow == watchWindow) ShowControl(whScroll);
                                                if (whichWindow == gameWindow)
                                                        EnableItem(myMenus[editM], undoCommand);
                                                else
                                                        DisableItem(myMenus[editM], undoCommand);
                                                TEFromScrap();
                                        }
                                        else
                                        {
                                                if (te) TEDeactivate(te);
                                                if (vs) HideControl(vs);
                                                if (whichWindow == srcWindow) HideControl(shScroll);
                                                if (whichWindow == watchWindow) HideControl(whScroll);
                                                ZeroScrap();
                                                TEToScrap();
                                        }
                                        if (ours(whichWindow))
                                                DrawGrowIcon(whichWindow);
                                }
                                break;
                                
                        case updateEvt:
                                if (ours(whichWindow = (WindowPtr)myEvent.message))
                                        UpdateWindow(whichWindow);
                                break;
                        
                        case kHighLevelEvent: {
                                OSErr   err;
                                
                                err = AEProcessAppleEvent(&myEvent);
                                assert(err == noErr);
                                break;
                        }
                        
                        default:
                                break;
                }
        }
        
        if (gsQuitting) {
                TypeCommand("quit", (tadsdebug || tdb2) ? dTEH : TEH);
                gsQuitting = false;
                // Reset gsQuitting so it doesn't interfere with getting the
                // user's response as to whether they really want to quit.
                return 0;
        }
        else
                return 1;
}

void runtimeMenuSetup(void)
{
        /* XXX: This function ought to be in os0tr_ma.c! */
        int                     i;
        
        if (tadsdebug || tdb2) return;
//      myMenus[appleM] = NewMenu(appleID, "\p\024");
//      AppendMenu(myMenus[appleM], "\pAbout TADS...");
//      AppendMenu(myMenus[appleM], "\p(-");
//      AppendResMenu(myMenus[appleM], 'DRVR');
        
        myMenus[appleM] = GetMenu(appleID);
        AppendResMenu(myMenus[appleM], 'DRVR');
        myMenus[fileM] = GetMenu(fileID);
        myMenus[editM] = GetMenu(editID);
        myMenus[cmdM] = GetMenu(cmdID);
        myMenus[travelM] = GetMenu(travelID);
        myMenus[fontM] = GetMenu(fontID);
        AppendResMenu(myMenus[fontM], 'FOND');
        myMenus[fontsizeM] = GetMenu(fontsizeID);
        
        for (i = appleM; i <= fontsizeM; i++)
                InsertMenu(myMenus[i], 0);
        DrawMenuBar();

}


static void SetGameWindowFont(ConstStr255Param fontName, short menuItem = 0)
{
        static short    oldItem = 0;
        static Str63    curFont = "\p";
        
        if (!StringEqual_pp(curFont, fontName)) {
                FontInfo                f;
                Point                   p;
                short                   fnum;
                ControlHandle   vs;
                TEHandle                te;
                short                   sizeItems;
                
                findinfo(gameWindow, &te, &vs);
                RTPSetFont(fontName);
                if (te) {
                        GetFNum(fontName, &fnum);
                        TextFont((*te)->txFont = fnum);
                        GetFontInfo(&f);
                        (*te)->lineHeight = f.ascent + f.descent;
                        (*te)->fontAscent = f.ascent;
                        p.v = p.h = 0;
                        MyGrowWindow(gameWindow, p);
                }

                // Update checks in font menu.
                if (oldItem > 0) CheckItem(myMenus[fontM], oldItem, false);
                if (menuItem == 0) {
                        for (short i = 1, lim = CountMItems(myMenus[fontM]); i <= lim; i++) {
                                Str255  itemText;
                                
                                GetMenuItemText(myMenus[fontM], i, itemText);
                                if (StringEqual_pp(fontName, itemText)) {
                                        menuItem = i;
                                        break;
                                }
                        }
                }
                if (menuItem > 0) CheckItem(myMenus[fontM], menuItem, true);
                oldItem = menuItem;
                
                // Outline real sizes in the size menu.
                sizeItems = CountMItems(myMenus[fontsizeM]);
                for (short i = 1; i < sizeItems - 3; i++) {
                        Str255  itemText;
                        long    size;
                        
                        GetMenuItemText(myMenus[fontsizeM], i, itemText);
                        StringToNum(itemText, &size);
                        SetItemStyle(myMenus[fontsizeM], i, (size > 0 && RealFont(fnum, size)) ? outline : normal);
                }
                SetItemStyle(myMenus[fontsizeM], sizeItems - 2, RealFont(fnum, 509) ? outline : normal);
        }
}

const short kMinimumFontSize = 6;
const short kMaximumFontSize = 48;

static void SetGameWindowFontSize(short size, short menuItem = 0)
{
        static short    oldItem = 0;
        static short    oldSize = 0;
        
        if (size < kMinimumFontSize) size = kMinimumFontSize;
        if (size > kMaximumFontSize) size = kMaximumFontSize;
        
        if (size != oldSize) {
                FontInfo                f;
                Point                   p;
                ControlHandle   vs;
                TEHandle                te;
                
                // Change the font size.
                findinfo(gameWindow, &te, &vs);
                RTPSetFontSize(size);
                if (te) {
                        TextSize((*te)->txSize = size);
                        GetFontInfo(&f);
                        (*te)->lineHeight = f.ascent + f.descent;
                        (*te)->fontAscent = f.ascent;
                        p.v = p.h = 0;
                        MyGrowWindow(gameWindow, p);
                }
                
                // Handle checking and unchecking items.
                if (oldItem > 0) CheckItem(myMenus[fontsizeM], oldItem, false);
                if (menuItem == 0) {
                        short           numItems = CountMItems(myMenus[fontsizeM]);
                        
                        menuItem = numItems - 2;  // Assume that it's the "Other" item.
                        for (short i = 1; i < numItems - 3; i++) {
                                Str255  itemText;
                                long    item;
                                
                                GetMenuItemText(myMenus[fontsizeM], i, itemText);
                                StringToNum(itemText, &item);
                                if (size == item) {
                                        menuItem = i;
                                        break;
                                }
                        }
                }
                if (menuItem > 0) CheckItem(myMenus[fontsizeM], menuItem, true);
                oldItem = menuItem;
        }
}

static bool GetOtherFontSize(short& newsize)
{
        DialogPtr       d;
        
        d = GetNewDialog(131, NULL, (WindowPtr) -1L);
        if (d) {
                const DialogItemIndex kSizeItem = 3;
                
                DialogItemIndex itemHit;
                Str255                  sizeText;
                long                    size;
                
                SetDialogDefaultItem(d, ok);
                SetDialogCancelItem(d, cancel);
                NumToString(RTPGetFontSize(), sizeText);
                SetControlItemTitle(d, kSizeItem, sizeText);
                SelectDialogItemText(d, kSizeItem, 0, 32767);
                ShowWindow(d);
                
                do {
                        ModalDialog(NULL, &itemHit);
                        switch (itemHit) {
                                case kSizeItem: {
                                        GetControlItemTitle(d, kSizeItem, sizeText);
                                        StringToNum(sizeText, &size);
                                        HiliteControlItem(d, ok, (size >= kMinimumFontSize
                                                        && size <= kMaximumFontSize) ? 0 : 255);
                                        break;
                                }
                        }
                } while (itemHit != ok && itemHit != cancel);
                
                if (itemHit == ok) {
                        GetControlItemTitle(d, kSizeItem, sizeText);
                        StringToNum(sizeText, &size);
                        newsize = size;
                }
                DisposeDialog(d);
                return (itemHit == ok);
        }
        else
                return false;
}

int DoCommand(long mResult)
{
        int                             theItem;
        Str255                  name;
        char               *newcmd = NULL;
        ControlHandle   vs;
        TEHandle                te;
        int                             mid;

        findinfo(myWindow, &te, &vs);
        theItem = LoWord(mResult);
        mid = HiWord(mResult);
        
        if (mid != appleID && tdb2 && (mid != editID || theItem > 7))
        {
                newcmd = dbgmenu(mResult);
                goto done;
        }
        
        switch (mid)
        {
                case appleID:
                        if (theItem == 1) /* about */
                        {
                                DialogPtr       mydlg;
                                short           itemHit;
                                
                                mydlg = GetNewDialog(130, NULL, (WindowPtr) -1L);
                                do
                                {
                                        ModalDialog(NULL, &itemHit);
                                } while (itemHit != 1);
                                DisposeDialog(mydlg);
                        }
                        else
                        {
                                GetMenuItemText(myMenus[appleM], theItem, name);
                                OpenDeskAcc(name);
                                SetPort(myWindow);
                        }
                        break;
                        
                case cmdID:
                        switch(theItem)
                        {
                                case cmInventory: newcmd = "inventory"; break;
                                case cmLook:  newcmd = "look"; break;
                                case cmScore: newcmd = "score"; break;
                                case cmWait:  newcmd = "wait"; break;
                                case cmTake:  newcmd = "take "; break;
                                case cmDrop:  newcmd = "drop "; break;
                        }
                        break;
                        
                case travelID:
                        switch(theItem)
                        {
                                case cmNorth: newcmd = "north"; break;
                                case cmSouth: newcmd = "south"; break;
                                case cmEast:  newcmd = "east"; break;
                                case cmWest:  newcmd = "west"; break;
                                case cmNE:    newcmd = "northeast"; break;
                                case cmNW:    newcmd = "northwest"; break;
                                case cmSE:    newcmd = "southeast"; break;
                                case cmSW:    newcmd = "southwest"; break;
                                case cmUp:    newcmd = "up"; break;
                                case cmDown:  newcmd = "down"; break;
                                case cmIn:    newcmd = "in"; break;
                                case cmOut:   newcmd = "out"; break;
                        }
                        break;
                        
                case fontID:
                        {
                                Str255                  name;
                                
                                GetMenuItemText(myMenus[fontM], theItem, name);
                                SetGameWindowFont(name, theItem);
                        }
                        break;
                
                case fontsizeID:
                        {
                                short           numMenuItems = CountMItems(myMenus[fontsizeM]);
                
                                if (theItem == numMenuItems)
                                        // Larger
                                        SetGameWindowFontSize(RTPGetFontSize() + 1);
                                else if (theItem == numMenuItems - 1)
                                        // Smaller
                                        SetGameWindowFontSize(RTPGetFontSize() - 1);
                                else if (theItem == numMenuItems - 2) {
                                        // Other...
                                        short   newSize;
                                        
                                        if (GetOtherFontSize(newSize))
                                                SetGameWindowFontSize(newSize);
                                }
                                else if (theItem == numMenuItems - 3)
                                        // ----
                                        ;
                                else {
                                        Str255  itemText;
                                        long    fontsize;
                                        
                                        GetMenuItemText(myMenus[fontsizeM], theItem, itemText);
                                        StringToNum(itemText, &fontsize);
                                        SetGameWindowFontSize(fontsize, theItem);
                                }
                        }
                        break;

                case fileID:
                        switch(theItem)
                        {
                                case fmQuit: newcmd = "quit"; break;  // XXX: To be truly clean, send self a "quit" event.
                                case fmSave: newcmd = "save"; break;
                                case fmRestore: newcmd = "restore"; break;
                                case fmRestart: newcmd = "restart"; break;
                        }
                        break;

                case editID:
                        if (SystemEdit(theItem-1) == 0)
                        {
                                findinfo(FrontWindow(), &te, &vs);
                                switch (theItem) 
                                {
                                        case undoCommand:
                                                newcmd = "undo";
                                                break;
                                                
                                        case cutCommand:
                                                if (FrontWindow() == myWindow && te) {
                                                        TECut(te);
                                                        if (ZeroScrap() == noErr)
                                                                (void) TEToScrap();
                                                }
                                                break;
                                                
                                        case copyCommand:
                                                if (te) {
                                                        OSErr   err;
                                                        
                                                        TECopy(te);
                                                        if ((err = ZeroScrap()) == noErr)
                                                                if ((err = TEToScrap()) == noErr)
                                                                        ;
                                                                else {
                                                                        SysBeep(1);
                                                                        DPRINTF(("TEToScrap returned error %d!", err));
                                                                }
                                                        else {
                                                                SysBeep(1);
                                                                DPRINTF(("ZeroScrap returned error %d!", err));
                                                        }
                                                }
                                                break;
                                                
                                        case pasteCommand:
                                                (void) TEFromScrap();
                                                if (FrontWindow() == myWindow && te
                                                        && (*te)->teLength - cmdpoint +  TEGetScrapLength() < 128)
                                                {
                                                        if ((*te)->selStart < cmdpoint)
                                                                TESetEmptySelection(te);
                                                        TEPaste(te);
                                                }
                                                break;
                                                
                                        case clearCommand:
                                                if (watchWindow && FrontWindow() == watchWindow)
                                                        dbgwxclr();
                                                else if (FrontWindow() == myWindow && te)
                                                        TEDelete(te);
                                                break;
                                }
                                ShowSelect();
                        }
                        break;
        }
        
done:
        HiliteMenu(0);
        
        if (newcmd) {
                TypeCommand(newcmd, te);
                return 0;
        }
        else
                return 1;
}

static int os_busy_flag;

void MaintainCursor(void)
{
        Point           pt;
        WindowPtr       wPtr;
        GrafPtr         savePort;
        TEHandle        te;
        
        if (os_busy_flag)
        {
                SetCursor(&waitCursor);
                return;
        }
        
        if (ours(wPtr = FrontWindow()))
        {
                te = (wPtr == dbgWindow ? dTEH : TEH);
                GetPort(&savePort);
                SetPort((GrafPtr)wPtr);
                GetMouse(&pt);
                if ((WindowPtr)wPtr == myWindow && PtInRect(pt, &(**te).viewRect))
                        SetCursor(&editCursor);
                else
                        SetCursor(&qdArrow);
                SetPort(savePort);
        }
}

void os_csr_busy(int f)
{
        os_busy_flag = f;
        SetCursor(f ? &waitCursor : &qdArrow);
}

void MaintainMenus(void)
{
        ControlHandle   vs;
        TEHandle                te;
        extern int              watch_selection;
        MenuHandle              em = myMenus[editM];
        
        findinfo(FrontWindow(), &te, &vs);
        if (!((WindowPeek)myWindow)->visible || !ours(FrontWindow()))
        {
                EnableItem(em, undoCommand);
                EnableItem(em, cutCommand);
                EnableItem(em, copyCommand);
                EnableItem(em, clearCommand);
        }
        else
        {
                if (FrontWindow() == gameWindow)
                        EnableItem(em, undoCommand);
                else
                                DisableItem(em, undoCommand);
                if (!te || (*te)->selStart==(*te)->selEnd)
                {
                        DisableItem(em, cutCommand);
                        DisableItem(em, copyCommand);
                        DisableItem(em, clearCommand);
                }
                else
                {
                        EnableItem(em, cutCommand);
                        EnableItem(em, copyCommand);
                        EnableItem(em, clearCommand);
                }
                if (te) EnableItem(em, pasteCommand);
                
                /*
                 *  Adjust menus based on whether selection is in the command line or not
                 */
                if (!te || (*te)->selStart < cmdpoint)
                {
                        DisableItem(em, cutCommand);
                /*      DisableItem(em, pasteCommand); */
                        DisableItem(em, clearCommand);
                }
                
                if (FrontWindow() != myWindow)
                {
                        DisableItem(em, pasteCommand);
                        DisableItem(em, cutCommand);
                        DisableItem(em, clearCommand);
                }
                
                if (FrontWindow() == watchWindow)
                {
                        if (watch_selection != -1)
                                EnableItem(em, clearCommand);
                }
        }
}

void SetUpCursors(void)
{
        CursHandle hCurs;
        
        hCurs = GetCursor(1);
        editCursor = **hCurs;
        hCurs = GetCursor(watchCursor);
        waitCursor = **hCurs;
}

bool ours(WindowPtr w)
{
        return(w && (w == gameWindow || w == dbgWindow || w == srcWindow  || w == watchWindow));
}

static TEHandle NewTextView(const Rect *destRect, const Rect *viewRect)
{
        TEHandle        te;
        
        te = TENew(destRect, viewRect);
        if (te) {
                TEAutoView(true, te);
        }
        return te;
}

void SetUpWindows(void)
{
        Rect    viewRect;
        Rect    vScrollRect;
        
        if (tadsdebug || tdb2) {
                myWindow = GetNewCWindow(windowID, &wRecord, (WindowPtr) -1L);
        }
        else {
                Rect    wRect;
                
                RTPGetWindowSize(wRect);
                myWindow = NewCWindow(&wRecord, &wRect, "\p", true, documentProc, (WindowPtr) -1L, false, 0L);
        }
        SetPort(myWindow);
        
        gameWindow = myWindow;
        TextFont(RTPGetFontID());
        TextSize(RTPGetFontSize());
        vScrollRect = myWindow->portRect;
        vScrollRect.left = vScrollRect.right - 15;
        vScrollRect.right += 1;
        vScrollRect.bottom -= 14;
        vScrollRect.top -= 1;
        vScroll = NewControl(myWindow, &vScrollRect, "\p", 1, 0, 0, 0, scrollBarProc, 0L);

        viewRect = qdPort->portRect;
        viewRect.right -= kScrollBarWidth;
        viewRect.bottom -= kScrollBarWidth;
        InsetRect(&viewRect, 4, 4);
        if (use_statline) viewRect.top += STATUS_OFFSET;
        TEH = NewTextView(&viewRect, &viewRect);
        SetView(qdPort, true);
                        
        // Check correct font menu items -- this sets pagelength and linewidth as a side effect.
        {
                Str63   fontName;
                
                RTPGetFontName(fontName);
                SetGameWindowFont(fontName);
                SetGameWindowFontSize(RTPGetFontSize());
        }
        
        if (tadsdebug || tdb2)
        {
                Rect            rect1;
                btndef     *p;
                FontInfo        f;
                extern int      dbgydelta;
                extern int      dbgpagesize;
                
                SetPort(dbgWindow = GetNewWindow(tdb2 ? 12075 : dbgWinID, &dbgWRecord, (WindowPtr) -1L));
                dbgWRecord.goAwayFlag = 0;
                
                vScrollRect = (*dbgWindow).portRect;

                if (tdb2)
                {
                        TextFont(4);
                        TextSize(9);
                        vScrollRect.top -= 1;
                }
                else
                {
                        vScrollRect.top += 22;
                }

                vScrollRect.left = vScrollRect.right - 15;
                vScrollRect.right += 1;
                vScrollRect.bottom -= 14;
                dvScroll = NewControl(dbgWindow, &vScrollRect, "\p", 1, 0, 0, 0, scrollBarProc, 0L);
                
                rect1.top = 3;
                rect1.bottom = 18;
                rect1.left = 3;
                rect1.right = 58;
                
                if (tdb2)
                {
                        viewRect = qdPort->portRect;
                        viewRect.right -= kScrollBarWidth;
                        viewRect.bottom -= kScrollBarWidth;
                        InsetRect(&viewRect, 4, 4);
                        dTEH = NewTextView(&viewRect, &viewRect);
                        SetView(qdPort, true);
                        TextFont(4);
                        TextSize(9);
                        
                        SetPort(srcWindow = GetNewWindow(19676, &srcWRecord, (WindowPtr) -1L));
                        
                        vScrollRect = srcWindow->portRect;
                        vScrollRect.top -= 1;
                        vScrollRect.left = vScrollRect.right - 15;
                        vScrollRect.right += 1;
                        vScrollRect.bottom -= 14;
                        svScroll = NewControl(srcWindow, &vScrollRect, "\p", 1, 0, 0, 0, scrollBarProc, 0L);
                        HideControl(svScroll);
                        
                        vScrollRect = srcWindow->portRect;
                        vScrollRect.top = vScrollRect.bottom - 15;
                        vScrollRect.bottom += 1;
                        vScrollRect.left -= 1;
                        vScrollRect.right -= 14;
                        shScroll = NewControl(srcWindow, &vScrollRect, "\p", 1, 0, 0, 1024, scrollBarProc, 0L);
                        HideControl(shScroll);
                }
                else
                { 
                        SetControlMaximum(dvScroll, 3);
                        SetControlMinimum(dvScroll, 1);
                        SetControlValue(dvScroll, 2);

                        for (p = btnlist ; p->btnname ; ++p)
                        {
                                p->btnctl = NewControl(dbgWindow, &rect1, p->btnname, 1, 0, 0, 0, 0, 0L);
                                (*p->btnctl)->contrlRfCon = p->btnrefcon;
                                rect1.left += 62;
                                rect1.right += 62;
                        }
                
                        GetFontInfo(&f);
                        dbgydelta = f.ascent + f.descent + f.leading;
                        dbgpagesize = (Dy(qdPort->portRect) - 24) / dbgydelta;
                }
        }
        
        SetPort(myWindow);
}


void AdjustText(WindowPtr w)
{
        if (w == srcWindow) {
                srcAdjust();
        }
        else {
                TEHandle                te = (w == dbgWindow ? dTEH : TEH);
                ControlHandle   vs = (w == dbgWindow ? dvScroll : vScroll);
                short                   oldScroll, newScroll, delta;
                
                oldScroll = (**te).viewRect.top - (**te).destRect.top;
                newScroll = GetControlValue(vs) * (**te).lineHeight;
                delta = oldScroll - newScroll;
                if (delta != 0)
                        TEScroll(0, delta, te);
        }
}

void SetControlEnd(void)
{
        TEHandle                te;
        ControlHandle   vs;
        int                             cval;
        int                             cmax;

        findinfo(myWindow, &te, &vs);
        if (!te) return;
        
        cmax = cval = (**te).nLines - linesInFolder;
        if ((*vs)->contrlValue == (*vs)->contrlMax)
        {
                /* no need to redraw - just set values directly */
                (*vs)->contrlValue = (*vs)->contrlMax = cmax;
        }
        else
        {
                (*vs)->contrlMax = cmax;
                SetControlValue(vs, cmax);
        }
}

void ShowEnd(void)
{
        TEHandle                te;
        ControlHandle   vs;
        int                             cval;
        int                             oldScroll;
        int                             newScroll;
        int                             delta;
        
        findinfo(myWindow, &te, &vs);
        if (!te) return;
        
        cval = (**te).nLines - linesInFolder;
        oldScroll = (**te).viewRect.top - (**te).destRect.top;
        newScroll = cval * (**te).lineHeight;
        delta = oldScroll - newScroll;
        if (cval > 0 && delta < 0)
                TEScroll(0, delta, te);
}

void SetVScroll(void)
{
        int                             n;
        TEHandle                te;
        ControlHandle   vs;
        
        findinfo(myWindow, &te, &vs);
        n = (**te).nLines - linesInFolder;

        if ((**te).teLength > 0 && (*((**te).hText))[(**te).teLength-1] == kCR)
                n++;

        SetControlMaximum(vs, n > 0 ? n : 0);
}

void ShowSelect(void)
{
        TEHandle                te;
        ControlHandle   vs;
        SInt16                  topLine, bottomLine, theLine;
        
        findinfo(myWindow, &te, &vs);
        if (!te) return;
        
        SetView(myWindow);
        SetVScroll();
        AdjustText(myWindow);
        
        topLine = GetControlValue(vs);
        bottomLine = topLine + linesInFolder;
        
        if ((**te).selStart < (**te).lineStarts[topLine]
                || (**te).selStart >= (**te).lineStarts[bottomLine])
        {
                for (theLine = 0;
                         theLine < (**te).nLines && (**te).selStart >= (**te).lineStarts[theLine];
                         theLine++)
                        ;
                SetControlValue(vs, theLine - linesInFolder / 2);
                AdjustText(myWindow);
        }
}

void SetView(WindowPtr w, bool recalcAll)
{
        TEHandle                te;
        ControlHandle   vs;

        findinfo(w, &te, &vs);
        if (te) {
                if (recalcAll) {
                        (**te).viewRect = w->portRect;
                        (**te).viewRect.right -= kScrollBarWidth;
                        (**te).viewRect.bottom -= kScrollBarWidth;
                        InsetRect(&(**te).viewRect, 4, 4);
                        if (use_statline && w != dbgWindow) (**te).viewRect.top += STATUS_OFFSET;
                }
                linesInFolder = Dy((**te).viewRect) / (**te).lineHeight;
                if (recalcAll) {
                        (**te).viewRect.bottom = (**te).viewRect.top + (**te).lineHeight*linesInFolder;
                        #ifndef NO_DEST_ADJUST
                        (**te).destRect.right = (**te).viewRect.right;
                        #endif /* NO_DEST_ADJUST */
                        TECalText(te);
                }
        }
}

        
void macedShostat(int doloc /* 1 if location is to be redrawn, 0 otherwise */)
{
        if (use_statline) {
                const short kStatusBarHeight = STATUS_OFFSET;
                
                extern Str255           statbuf;
                extern Str63            scorebuf;
                
                Rect                            statusBar, updateRect;
                TPortState                      portState(gameWindow);
                TFontState                      fontState;
                TFractEnableState       fractState(false);
                        // turn off fractional spacing if using Charcoal, since it
                        // looks like crap with it on.
        
                TextFont(0);
                TextSize(0);
                
                /*
                 *   Erase the rectangle occupied by the status line, not counting the lines below
                 */
                statusBar.top = gameWindow->portRect.top;
                statusBar.left = gameWindow->portRect.left;
                statusBar.right = gameWindow->portRect.right - kScrollBarWidth;
                statusBar.bottom = statusBar.top + kStatusBarHeight;
                
                updateRect = statusBar;
                updateRect.bottom -= 3;  // Don't include double lines.
                if (!doloc) updateRect.left = updateRect.right - (StringWidth(scorebuf) + 8);  // Only draw score (maybe).
                EraseRect(&updateRect);
                
                /*
                 *   Draw the location string
                 */
                if (doloc)
                {
                        MoveTo(statusBar.left + 2, statusBar.bottom - 6);
                        DrawString(statbuf);
                }
                
                /*
                 *   Draw the score string
                 */
                MoveTo(statusBar.right - StringWidth(scorebuf) - 2, statusBar.bottom - 6);
                DrawString(scorebuf);
                
                /*
                 *   Draw the double line below it all
                 */
                MoveTo(statusBar.left, statusBar.bottom - 1);
                LineTo(statusBar.right, statusBar.bottom - 1);
                MoveTo(statusBar.left, statusBar.bottom - 3);
                LineTo(statusBar.right, statusBar.bottom - 3);
        }
}

void UpdateWindow(WindowPtr theWindow)
{
        GrafPtr savePort;
        
        GetPort(&savePort);
        SetPort(theWindow);
        BeginUpdate(theWindow);
        EraseRect(&theWindow->portRect);
        
        if (theWindow == dbgWindow)
        {
                if (tdb2)
                {
                        TEUpdate(&theWindow->portRect, dTEH);
                        DrawGrowIcon(theWindow);
                }
                else
                {
                        MoveTo(0, 20);
                        LineTo(Dx(qdPort->portRect), 20);
                        MoveTo(0, 22);
                        LineTo(Dx(qdPort->portRect), 22);
                        macedDbUpd();
                        MoveTo(qdPort->portRect.right - 15, qdPort->portRect.bottom - 15);
                        LineTo(qdPort->portRect.right - 15, qdPort->portRect.bottom);
                }
        }
        else if (theWindow == srcWindow)
        {
                dbgsrcupd();
                DrawGrowIcon(theWindow);
        }
        else if (theWindow == watchWindow)
        {
                dbgwatchupd();
                DrawGrowIcon(theWindow);
        }
        else
        {
                macedShostat(1);
                TEUpdate(&theWindow->portRect, TEH);
                DrawGrowIcon(theWindow);
        }
        
        /* draw controls */
        DrawControls(theWindow);

        EndUpdate(theWindow);
        SetPort(savePort);
}




static pascal void ScrollProc(ControlHandle theControl, short theCode)
{
        int                     pageSize;
        int                     scrollAmt;
        TEHandle        te;
        
        if (theControl == vScroll) te = TEH;
        else if (theControl == dvScroll) te = dTEH;
        else if (theControl == svScroll || theControl == shScroll)
        {
                srcScroll(theControl, theCode);
                return;
        }
        else if (theControl == whScroll || theControl == wvScroll)
        {
                dbgwxScroll(theControl, theCode);
                return;
        }
        
        if (theCode == 0) return;
        
        pageSize = Dy((**te).viewRect) / (**te).lineHeight - 1;
                                        
        switch (theCode)
        {
                case kControlUpButtonPart:
                        scrollAmt = -1;
                        break;
                case kControlDownButtonPart:
                        scrollAmt = 1;
                        break;
                case kControlPageUpPart:
                        scrollAmt = -pageSize;
                        break;
                case kControlPageDownPart:
                        scrollAmt = pageSize;
                        break;
        }
        SetControlValue(theControl, GetControlValue(theControl) + scrollAmt);
        AdjustText(theControl == vScroll ? gameWindow : dbgWindow);
}

static int doButton(ControlHandle control)
{
        extern char onechar;
        
        if (onechar) {
                onechar = (*control)->contrlRfCon;
                return 0;
        }
        else
                return 1;
}

int DoContent(WindowPtr theWindow, EventRecord *theEvent)
{
        int                             cntlCode;
        ControlHandle   theControl;
        GrafPtr                 savePort;
        int                             okToTrack;
        
        if (tdb2)
                okToTrack = ours(theWindow);
        else
                okToTrack = (theWindow == myWindow);
        
        GetPort(&savePort);
        SetPort(theWindow);
        GlobalToLocal(&theEvent->where);
        if ((cntlCode = FindControl(theEvent->where, theWindow, &theControl)) == 0)
        {
                if (theWindow == srcWindow)
                {
                        dbgContent(theEvent->where, theEvent);
                }
                else if (theWindow == dbgWindow)
                {
                        if (tdb2 && PtInRect(theEvent->where, &(**dTEH).viewRect))
                        {
                                TEClick(theEvent->where, (theEvent->modifiers & shiftKey) != 0, dTEH);
                                if ((*dTEH)->selStart < cmdpoint && (*dTEH)->selStart == (*dTEH)->selEnd)
                                        TESetEmptySelection(dTEH);
                        }
                }
                else if (theWindow == watchWindow)
                {
                        dbgwxcont(theEvent->where, theEvent);
                }
                else if (PtInRect(theEvent->where, &(**TEH).viewRect))
                {
                        TEClick(theEvent->where, (theEvent->modifiers & shiftKey)!=0, TEH);
                        if ((*TEH)->selStart < cmdpoint && (*TEH)->selStart == (*TEH)->selEnd)
                                TESetEmptySelection(TEH);
                }
        }
        else if (cntlCode == kControlIndicatorPart && okToTrack)
        {
                TrackControl(theControl, theEvent->where, 0L);
                if (theControl == svScroll || theControl == shScroll)
                {
                        srcAdjust();
                        dbginval();
                }
                else if (theControl == wvScroll || theControl == whScroll)
                        dbgwxinval();
                else
                        AdjustText(theWindow);
        }
        else if ((cntlCode == kControlUpButtonPart || cntlCode == kControlPageUpPart) 
                        && theWindow == dbgWindow && onechar)
        {
                SetPort(savePort);
                onechar = '-';
                return(0);
        }
        else if ((cntlCode == kControlDownButtonPart || cntlCode == kControlPageDownPart)
                        && theWindow == dbgWindow && onechar)
        {
                SetPort(savePort);
                onechar = '+';
                return(0);
        }
        else if (cntlCode == kControlButtonPart)
        {
                if (TrackControl(theControl, theEvent->where, 0L))
                {
                        SetPort(savePort);
                        return(doButton(theControl));
                }
        }
        else if (okToTrack) {
                ControlActionUPP        scrollprocUPP;
                
                scrollprocUPP = NewControlActionProc(ScrollProc);
                if (scrollprocUPP) {
                        TrackControl(theControl, theEvent->where, scrollprocUPP);
                        DisposeRoutineDescriptor(scrollprocUPP);
                }
        }

        SetPort(savePort);
        return(1);
}

void MyGrowWindow(WindowPtr w, Point p)
{
        GrafPtr                 savePort;
        long                    theResult;
        Rect                    r;
        TEHandle                te;
        ControlHandle   vs;
        
        findinfo(w, &te, &vs);
        GetPort(&savePort);
        SetPort(w);
        
        if (p.v || p.h) {
                SetRect(&r, 80, 80, qd.screenBits.bounds.right, qd.screenBits.bounds.bottom);
                theResult = GrowWindow(w, p, &r);
                if (theResult == 0) {
                        /* Nothing changed; leave now. */
                        goto L_exit;
                }
                SizeWindow(w, LoWord(theResult), HiWord(theResult), 1);
                
                if (w == gameWindow) {
                        TPortState      portState(w);
                        Rect            bounds;
                        
                        bounds = w->portRect;
                        LocalToGlobal((Point *) &bounds.top);
                        LocalToGlobal((Point *) &bounds.bottom);
                        RTPSetWindowUserSize(bounds);
                }
        }
        
        InvalRect(&w->portRect);
        SetView(w, true);
        HidePen();

        /* resize the vertical scroll bar */
        if (vs)
        {
                MoveControl(vs, w->portRect.right - kScrollBarWidth, w->portRect.top - 1);
                SizeControl(vs, kScrollBarWidth + 1, Dy(w->portRect) - (kScrollBarWidth - 2));
        }
        
        /* resize the horizontal scroll bar, too, if present */
        if (w == srcWindow)
        {
                MoveControl(shScroll, w->portRect.left - 1, w->portRect.bottom - kScrollBarWidth);
                SizeControl(shScroll, Dx(w->portRect) - (kScrollBarWidth - 2), kScrollBarWidth + 1);
        }
        
        if (w == watchWindow)
        {
                MoveControl(whScroll, w->portRect.left - 1, w->portRect.bottom - kScrollBarWidth);
                SizeControl(whScroll, Dx(w->portRect) - (kScrollBarWidth - 2), kScrollBarWidth + 1);
                dbgwxgrow();
        }

        ShowPen();

        if (te && w == myWindow)
        {
                extern int      pagelength, linewidth;
                
                SetVScroll();
                AdjustText(w);
                pagelength = ((Dy(w->portRect) - (kScrollBarWidth-2)) / (*te)->lineHeight) - 4;
                linewidth = (Dx(w->portRect) - kScrollBarWidth - 8) / CharWidth('_');
        }
        
L_exit:
        SetPort(savePort);
}


void os_sleep_ms(long delay_in_milliseconds)
{
        UInt32          delay_in_ticks = (UInt32) (delay_in_milliseconds / 1000.0 * 60.0);
        UInt32          ticks;
        
        Delay(delay_in_ticks, &ticks);
}

long os_get_sys_clock_ms(void)
{
        return (long) (TickCount() * (1000.0 / 60.0));
        // XXX: This will give weird results if the system is up for more than 24 days!
}

int os_get_event(unsigned long timeout_in_milliseconds, int use_timeout,
                 os_event_info_t *info)
{
        Boolean         gotEvent;
        EventRecord     ev;
        UInt32          sleepTicks;
        
        if (use_timeout)
                sleepTicks = timeout_in_milliseconds * 0.06;
        else
                sleepTicks = -1;
        
        gotEvent = WaitNextEvent(keyDownMask, &ev, (UInt32) (timeout_in_milliseconds * 0.06), NULL);
        if (!gotEvent || ev.what == nullEvent) {
                return OS_EVT_TIMEOUT;
        }
        else {
                assert(ev.what == keyDown);
                info->key[0] = ev.message & charCodeMask;
                info->key[1] = 0;
                return OS_EVT_KEY;
        }
}

void os_set_title(const char *title)
{
        setwtitle(gameWindow, title);
}
