/*
==========================================================================================

	File:		console.c

	Contains:	Provides printf to a disk file.

	Copyright:	Based on Metrowerks code (see below);
				modifications by Christopher Nebel.

	Change History (most recent first):
	
		<1>		 8/25/98	Neb		Initial check-in.

==========================================================================================
*/


/*  Metrowerks Standard Library  Version 2.1.2  1997 May  */

/*****************************************************************************/
/*	Project...:	Standard ANSI-C Library									     */
/*	Name......:	console.c												     */
/*	Purpose...:	Stubs for console.c										     */
/*  Copyright.: Copyright  1994-1997 Metrowerks, Inc. All rights reserved.  */
/*****************************************************************************/

#ifndef __CONSOLE__
#include <console.h>
#endif


#include <stdio.h>
#include <assert.h>

static char *fstdoutname = "stdout";
static FILE *fstdout;

/*
 *	The following four functions provide the UI for the console package.
 *	Users wishing to replace SIOUX with their own console package need
 *	only provide the four functions below in a library.
 */

/*
 *	extern short InstallConsole(short fd);
 *
 *	Installs the Console package, this function will be called right
 *	before any read or write to one of the standard streams.
 *
 *	short fd:		The stream which we are reading/writing to/from.
 *	returns short:	0 no error occurred, anything else error.
 */

short InstallConsole(short fd)
{
//	#pragma unused (fd)
	
	assert(fstdout == NULL);
	
	fstdout = fopen(fstdoutname, "w+");	// Open for buffered writing.

	return (fstdout == NULL) ? 1 : 0;
}

/*
 *	extern void RemoveConsole(void);
 *
 *	Removes the console package.  It is called after all other streams
 *	are closed and exit functions (installed by either atexit or _atexit)
 *	have been called.  Since there is no way to recover from an error,
 *	this function doesn't need to return any.
 */

void RemoveConsole(void)
{
	if (fstdout) {
		fclose(fstdout);
		fstdout = NULL;
	}
}

/*
 *	extern long WriteCharsToConsole(char *buffer, long n);
 *
 *	Writes a stream of output to the Console window.  This function is
 *	called by write.
 *
 *	char *buffer:	Pointer to the buffer to be written.
 *	long n:			The length of the buffer to be written.
 *	returns short:	Actual number of characters written to the stream,
 *					-1 if an error occurred.
 */

long WriteCharsToConsole(char *buffer, long n)
{
//	#pragma unused (buffer, n)
	
	size_t written;
	
	written = fwrite(buffer, 1, n, fstdout);

	return (written == n) ? written : -1;
}

/*
 *	extern long ReadCharsFromConsole(char *buffer, long n);
 *
 *	Reads from the Console into a buffer.  This function is called by
 *	read.
 *
 *	char *buffer:	Pointer to the buffer which will recieve the input.
 *	long n:			The maximum amount of characters to be read (size of
 *					buffer).
 *	returns short:	Actual number of characters read from the stream,
 *					-1 if an error occurred.
 */

long ReadCharsFromConsole(char *buffer, long n)
{
	#pragma unused (buffer, n)

	return 0;
}

/*
 *	extern char *__ttyname(long fildes);
 *
 *	Return the name of the current terminal (only valid terminals are
 *	the standard stream (ie stdin, stdout, stderr).
 *
 *	long fildes:	The stream to query.
 *
 *	returns char*:	A pointer to static global data which contains a C string
 *					or NULL if the stream is not valid.
 */

extern char *__ttyname(long fildes)
{
	static char *__devicename = "null device";

	if (fildes == 0)
		return (fstdoutname);
	else if (fildes >= 0 && fildes <= 2)
		return (__devicename);
	else
		return NULL;
}

/*     Change record
*/
